#include <sys/socket.h> /* for socket()    */
#include <stdio.h>      /* for perror()    */
#include <string.h>     /* for memset()    */
#include <unistd.h>     /* for sleep()     */

#include "socket.hh"

void Socket::BecomeListenerTCP(int port)
{
    int sock, RetryCount=0;
    
RetrySocket:
    sock = socket(AF_INET, SOCK_STREAM, 0);
    if(sock < 0)
    {
        perror("socket");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetrySocket;
    }

    struct sockaddr_in mysock;

    memset(&mysock, 0, sizeof(mysock));

    mysock.sin_family = AF_INET;
    mysock.sin_addr.s_addr = htonl(INADDR_ANY);
    /* FIXME: Allow this to be selected^ */
    mysock.sin_port = htons(port);
    
    int dummy = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy);
    
RetryBind:
    if(bind(sock, (struct sockaddr *)&mysock, sizeof(mysock)) < 0)
    {
        perror("bind");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryBind;
    }
    
    listen(sock, 5);
    
    Type          = Socket::SockTypeListening;
    Protocol      = SockProtocolTCP;
    fd            = sock;
    Peer.DestPort = port;
}

void Socket::BecomeListenerUDP(int port)
{
    int sock, RetryCount=0;
    
RetrySocket:
    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if(sock < 0)
    {
        perror("socket");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetrySocket;
    }

    struct sockaddr_in mysock;

    memset(&mysock, 0, sizeof(mysock));

    mysock.sin_family = AF_INET;
    mysock.sin_addr.s_addr = htonl(INADDR_ANY);
    /* FIXME: Allow this to be selected^ */
    mysock.sin_port = htons(port);
    
    int dummy = 1;
    setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &dummy, sizeof dummy);
    
RetryBind:
    if(bind(sock, (struct sockaddr *)&mysock, sizeof(mysock)) < 0)
    {
        perror("bind");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryBind;
    }
    
    Type          = Socket::SockTypeListening;
    Protocol      = SockProtocolUDP;
    fd            = sock;
    Peer.DestPort = port;
}

DataPacket Socket::AcceptTCP(const Socket &parent)
{
    int RetryCount = 0;
    
    Peer = parent.Peer;
    
RetryAccept:
    Peer.AddrLen = sizeof(Peer.Addr);

    int sock = accept(parent.fd, (struct sockaddr *)&Peer.Addr, &Peer.AddrLen);
    if(sock < 0)
    {
        perror("accept");
        /* FIXME: Now what? */
        sleep(1);
        if(++RetryCount<50)goto RetryAccept;
    }
    
    Type     = SockTypeConnected;
    Protocol = SockProtocolTCP;
    fd       = sock;
    
    DataPacket pak;
    pak.Peer = Peer;
    pak.Type = DataPacket::PacketTypeOpen;
    return pak;
}

DataPacket Socket::ReceiveUDP()
{
    Datatype Buffer[16384];
    DataPacket pak;
    
    Peer.AddrLen = sizeof(Peer.Addr);

    int Size = recvfrom(fd, Buffer, sizeof(Buffer), 0, (struct sockaddr *)&Peer.Addr, &Peer.AddrLen);
    if(Size < 0)
    {
        perror("recvfrom");
        /* FIXME: What now? */
        return pak;
    }
    
    pak.Peer = Peer;
    pak.Type = DataPacket::PacketTypeData;
    pak.MakeData(Buffer, Size);
    return pak;
}

DataPacket Socket::ReceiveTCP()
{
    Datatype Buffer[16384];
    DataPacket pak;
    
    int Size = read(fd, Buffer, sizeof(Buffer));
    if(Size < 0)
    {
        perror("read");
        /* FIXME: What now? */
        return pak;
    }
    if(!Size)
    {
        /* EOF. Ignore it. */
        return pak;
    }
    
    pak.Peer = Peer;
    pak.Type = DataPacket::PacketTypeData;
    pak.MakeData(Buffer, Size);
    return pak;
}

DataPacket Socket::Die()
{
    close(fd);
    
    DataPacket pak;
    
    if(Type == SockTypeConnected)
    {
        pak.Peer = Peer;
        pak.Type = DataPacket::PacketTypeClose;
        return pak;
    }
}

void Socket::SendTCP()
{
    const Datatype *buf = SendQueue.GetBegin();
    unsigned        len = SendQueue.GetSize();
    
    int status = write(fd, buf, len);
    if(status < 0)
    {
        /* FIXME: What should we do instead? */
        perror("write");
        return;
    }
    
    if(status > 0)
        SendQueue.Eat(status);
}

void Socket::SendUDP()
{
    const Datatype *buf = SendQueue.GetBegin();
    unsigned        len = SendQueue.GetSize();
    
    int status = sendto(fd, buf, len, 0, &Peer.Addr, &Peer.AddrLen);
    if(status < 0)
    {
        /* FIXME: What should we do instead? */
        perror("sendto");
        return;
    }
    
    if(status > 0)
        SendQueue.Eat(status);
}
